/*

    module to define and use macros within apache runtime configuration files.
    Copyright (C) 1998 Fabien COELHO

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

 */

/*
  $Id: mod_macro.c,v 1.30 1998/12/19 17:05:35 coelho Exp $

  $Log: mod_macro.c,v $
  Revision 1.30  1998/12/19 17:05:35  coelho
  many fixes to improve error and warning messages.
  in particular, everything should be as clear as possible
  regarding error localisation. however it must not be too verbose.

  Revision 1.29  1998/12/19 15:14:09  coelho
  style++.
  some comments.
  simpler say_it implementation.

  Revision 1.28  1998/12/19 12:05:53  coelho
  more apache style.
  'const' added to some arguments.

  Revision 1.27  1998/12/19 10:17:13  coelho
  typo.

  Revision 1.26  1998/12/19 10:16:23  coelho
  soft abort of configuration file processing under Error.

  Revision 1.25  1998/12/19 10:10:41  coelho
  Error generates a real error instead of a syntax error.

  Revision 1.24  1998/12/19 09:57:16  coelho
  Error and Warning stuff added.

  Revision 1.23  1998/12/18 18:56:41  coelho
  warn about empty arguments when used...

  Revision 1.22  1998/12/18 18:46:17  coelho
  error on empty argument names.

  Revision 1.21  1998/12/18 18:33:39  coelho
  small style fixes.

  Revision 1.20  1998/12/18 17:36:20  coelho
  check macro contents for unsused arguments.

  Revision 1.19  1998/12/18 17:02:27  coelho
  small style fixes.
  function order changed.

  Revision 1.18  1998/12/18 16:36:15  coelho
  simpler code.

  Revision 1.17  1998/12/18 16:29:14  coelho
  small style fixes.

  Revision 1.16  1998/12/18 15:53:47  coelho
  small style fix.

  Revision 1.15  1998/12/18 15:52:32  coelho
  better error messages (macro names are quoted).
  style++.

  Revision 1.14  1998/12/18 15:40:14  coelho
  untabbed.

  Revision 1.13  1998/12/18 15:38:51  coelho
  I changed my mind. chose longest argument instead of first.
  Thus no more conflicts if argument names are different.
  Warn anyway in case of argument name prefix conflict.
  Error if the same argument name appears more than once.

  Revision 1.12  1998/12/18 14:40:47  coelho
  warn the user about potential argument name prefix conflicts.
  how nice I am!

  Revision 1.11  1998/12/18 12:32:06  coelho
  small style fix.

  Revision 1.10  1998/12/18 10:24:02  coelho
  style fixes to match "Apache Developpers' C Language Style Guide" closely.

  Revision 1.9  1998/12/18 08:02:25  coelho
  comments++

  Revision 1.8  1998/12/18 08:00:25  coelho
  cosmetics and comments.

  Revision 1.7  1998/12/17 09:09:39  coelho
  Macro and Use can appear anywhere (?).
  Check for badly nested limit sections.
  To avoid compile-time warnings global 'macros' -> 'all_macros'.
  Fixed end_token to control nesting after expansion.

  Revision 1.6  1998/12/16 22:19:16  coelho
  cosmetics...

  Revision 1.5  1998/12/16 18:27:33  coelho
  typo.

  Revision 1.4  1998/12/16 17:50:01  coelho
  simpler.
  allow redefinitions with a warning instead of issuing an error.

  Revision 1.3  1998/12/16 14:22:18  coelho
  moved to GNU GPL v2.
  Discussion moved to index.html.

  Revision 1.2  1998/12/16 12:58:42  coelho
  case-insensitive comparisons.

  Revision 1.1  1998/12/15 18:06:41  coelho
  Initial revision

*/

#include "httpd.h"
#include "http_config.h"
#include "http_log.h"

/************************************************ COMPILE TIME DEBUG CONTROL */
/* #define DEBUG_MODULE_MACRO 1
   run -f /users/cri/coelho/Misc/apache/mod_macro/Tests/test??.conf
*/

#if defined(debug)
#undef debug
#endif

#if defined(DEBUG_MODULE_MACRO)
#define debug(stmt) stmt
#else
#define debug(stmt)
#endif

/********************************************************** MACRO MANAGEMENT */

module macro_module;

/* a macro: name, arguments, contents...
 */
typedef struct {
    char * name; /* case-insensitive name of the macro. */
    array_header /* of char* */ * arguments;
    array_header /* of char* */ * lines;
    int processing; /* whether the macro is being processed. */
    char * location; /* of the macro definition. */
} macro_t;

/* configuration tokens.
 */
#define BEGIN_MACRO "<Macro"
#define END_MACRO "</Macro>"
#define USE_MACRO "Use"

#define empty_string_p(p) (!(p) || *(p) == '\0')

/* macros are kept globally...
   they are not per-server or per-directory entities.

   I would need a hook BEFORE and AFTER configuration processing
   to initialize and close them properly.

   I would have such a hook if in the main/... I guess.

   The "initializer" does not seem to be called before.

   HACK:
   -----
   I put them in the temp_pool.
   restarts are detected because the temp_pool has changed...
   note that there is always a restart implicitely to check for the syntax.
*/
static array_header * all_macros = NULL;

/* returns the macro structure for name, or NULL if not found.
 */
static macro_t * get_macro_by_name(const array_header * macros,
                                   const char * name)
{
    int i;
    macro_t ** tab;
    ap_assert(macros);
    tab = (macro_t **)macros->elts;

    for (i = 0; i < macros->nelts; i++) {
        if (!strcasecmp(name, tab[i]->name)) {
            return tab[i];
        }
    }
    return NULL;
}

/* unmark recursion flags (used to detect macro recursion errors).
 */
static void clean_array_of_macros_recursion_flag(array_header * macros)
{
    int i;
    macro_t ** tab;
    ap_assert(macros);
    tab = (macro_t **)macros->elts;

    for (i = 0; i < macros->nelts; i++) {
        tab[i]->processing = 0;
    }
}

/* configuration state initialization.
   the state is simply an array_header which holds the macros.
 */
static void macro_init(pool * p)
{
    static ap_pool * last_time_temp_pool_hack = NULL;

    debug(fprintf(stderr, "macro_init\n"));

    /* is it a restart? what about concurrent threads?
     */
    if (last_time_temp_pool_hack != p) {
        last_time_temp_pool_hack = p;
        all_macros = ap_make_array(p, 1, sizeof(macro_t *));
        debug(fprintf(stderr, "macro_init done for %p\n", p));
    }
}

/*************************************************************** PARSE UTILS */

#define trim(line) while (*(line)==' ' || *(line)=='\t') (line)++

/* return configuration-parsed arguments from line as an array.
   the line is expected not to contain any '\n'?
 */
static array_header * get_arguments(ap_pool * p, const char * line)
{
    array_header * args = ap_make_array(p, 1, sizeof(char *));
    char * arg, ** new;

    trim(line);
    while (*line) {
        arg = ap_getword_conf(p, &line);
        new = ap_push_array(args);
        *new = arg;
        trim(line);
    }

    return args;
}

/* get read lines as an array till end_token.
   counts nesting for begin_token/end_token.
   it assumes a line-per-line configuration (thru getline).
   this function could be exported.
   begin_token may be NULL.
*/
static char * get_lines_till_end_token(ap_pool * p,
                                       configfile_t * config_file,
                                       const char * end_token,
                                       const char * begin_token,
                                       array_header ** plines)
{
    array_header * lines = ap_make_array(p, 1, sizeof(char *));
    char ** new, * first, * ptr;
    char line[MAX_STRING_LEN]; /* sorry, but that is expected by getline. */
    int nesting = 1;

    while (!ap_cfg_getline(line, MAX_STRING_LEN, config_file)) {
        ptr = line;
        first = ap_getword_conf_nc(p, &ptr);
        if (first) {
            if (!strcasecmp(first, end_token)) { /* okay! */
                nesting--;
                if (!nesting) {
                    *plines = lines;
                    return NULL;
                }
            }
            else if (begin_token && !strcasecmp(first, begin_token)) {
                nesting++;
            }
        }
        /* free first. */
        new  = ap_push_array(lines);
        *new = ap_psprintf(p, "%s\n", line); /* put '\n' back */
    }

    return ap_psprintf(p, "expected token not found: %s", end_token);
}

/* generates an error on macro with two arguments of the same name.
   generates an error if a macro argument name is empty.
   generates a warning if arguments name prefixes conflict.
*/
static const char * check_macro_arguments(ap_pool * p, const macro_t * macro)
{
    char ** tab = (char **)macro->arguments->elts;
    int nelts = macro->arguments->nelts, i, j;
    size_t ltabi, ltabj;

    for (i = 0; i < nelts; i++) {
        ltabi = strlen(tab[i]);

        if (ltabi == 0) {
            return ap_psprintf(p,
                               "macro '%s' (%s)\n\tempty argument #%d name\n",
                               macro->name, macro->location, i + 1);
        }

        for (j = i + 1; j < nelts; j++) {
            ltabj = strlen(tab[j]);

            if (!strcmp(tab[i], tab[j])) {
                return ap_psprintf(p,
                                  "argument name conflict in macro '%s' (%s)\n"
                                   "\targument %s: #%d and #%d\n"
                                   "\tchange argument names!",
                                   macro->name, macro->location,
                                   tab[i], i + 1, j + 1);
            }

            if (!strncmp(tab[i], tab[j], ltabi < ltabj ? ltabi : ltabj)) {
                ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_WARNING, NULL,
                        "macro '%s' (%s)\n"
                        "\targument name prefix conflict (%s #%d and %s #%d)\n"
                        "\tbe careful about your macro definition!\n",
                             macro->name, macro->location,
                             tab[i], i + 1, tab[j], j + 1);
            }
        }
    }

    return NULL;
}

/* warn about empty strings in array.
 */
static void check_macro_use_arguments(const char * where,
                                      const array_header * array)
{
    int i;
    char ** tab = (char **)array->elts;

    for (i = 0; i < array->nelts; i++) {
        if (empty_string_p(tab[i])) {
            ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_WARNING, NULL,
                         "%s\n\tempty argument #%d\n", where, i + 1);
        }
    }
}

/******************************************************** SUBSTITUTION UTILS */

/* replace name by replacement at the beginning of buf of bufsize.
   returns an error message or NULL.
 */
static char * substitute(char * buf, int bufsize,
                         const char * name, const char * replacement)
{
    char tmp[MAX_STRING_LEN]; /* sorry. should I malloc and free? */
    int lbuf = strlen(buf),
        lname = strlen(name),
        lrepl = strlen(replacement);

    if (lbuf + lrepl - lname >= bufsize) {
        return "cannot substitute, buffer size too small";
    }
    else if (lbuf + lrepl - lname >= MAX_STRING_LEN) {
        return "cannot substitute, tmp size too small";
    }

    /* simple. */
    strcpy(tmp, replacement);
    strcpy(tmp + lrepl, buf + lname);
    strcpy(buf, tmp);

    return NULL;
}

/* find first occurence of args in buf.
   in case of conflict, the LONGEST argument is kept. (could be the FIRST?).
   returns the pointer and the whichone found, or NULL.
 */
static char * next_substitution(const char * buf,
                                const array_header * args, int *whichone)
{
    int i;
    char * first = NULL, * found, ** tab = (char **)args->elts;
    size_t lfirst = 0, lfound;

    for (i=0; i<args->nelts; i++) {
        found = strstr(buf, tab[i]);
        lfound = strlen(tab[i]);
        if (found &&
            (!first || found < first || (found == first && lfirst < lfound))) {
            first = found;
            lfirst = lfound;
            *whichone = i;
        }
    }

    return first;
}

/* substitute macro arguments by replacements in buf of bufsize.
   returns an error message or NULL.
   if used is defined, returns the used macro arguments.
*/
static char * substitute_macro_args(char * buf, int bufsize,
                                    const macro_t * macro,
                                    const array_header * replacements,
                                    array_header * used)
{
    char * ptr = buf, * errmsg,
        ** atab = (char **)macro->arguments->elts,
        ** rtab = (char **)replacements->elts;
    int whichone;

    if (used) {
        ap_assert(used->nalloc >= replacements->nelts);
    }

    while ((ptr = next_substitution(ptr, macro->arguments, &whichone))) {
        errmsg = substitute(ptr, buf - ptr + bufsize,
                            atab[whichone], rtab[whichone]);
        if (errmsg) {
            return errmsg;
        }
        ptr += strlen(rtab[whichone]);
        if (used) {
            used->elts[whichone] = 1;
        }
    }

    return NULL;
}

/* perform substitutions in a macro contents and
   return the result as a newly allocated array, if result is defined.
   may also return an error message.
   passes used down to substitute_macro_args.
*/
static const char * process_content(ap_pool * p,
                                    const macro_t * macro,
                                    const array_header * replacements,
                                    array_header * used,
                                    array_header ** result)
{
    array_header * lines = macro->lines;
    char ** new, * errmsg, line[MAX_STRING_LEN]; /* sorry again. */
    int i;

    if (result) {
        *result = ap_make_array(p, 1, sizeof(char *));
    }

    for (i = 0; i < lines->nelts; i++) {
        strncpy(line, ((char **)lines->elts)[i], MAX_STRING_LEN - 1);
        errmsg = substitute_macro_args(line, MAX_STRING_LEN,
                                       macro, replacements, used);
        if (errmsg) {
            return ap_psprintf(p, "while processing line %d of macro '%s'"
                               " (%s)\n\t%s",
                               i + 1, macro->name, macro->location, errmsg);
        }

        if (result) {
            new = ap_push_array(*result);
            *new = ap_pstrdup(p, line);
        }
    }

    return NULL;
}

/* warn if some macro arguments are not used.
 */
static const char * check_macro_contents(ap_pool * p, const macro_t * macro)
{
    int nelts = macro->arguments->nelts, i;
    array_header * used = ap_make_array(p, nelts, sizeof(char));
    const char * errmsg;
    char ** names = (char **)macro->arguments->elts;

    for (i = 0; i < nelts; i++) {
        used->elts[i] = 0;
    }

    errmsg = process_content(p, macro, macro->arguments, used, NULL);

    if (errmsg) {
        /* free used. */
        return errmsg;
    }

    for (i = 0; i < nelts; i++) {
        if (!used->elts[i]) {
            ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_WARNING, NULL,
                         "macro '%s' (%s)\n\targument %s (#%d) never used\n",
                         macro->name, macro->location, names[i], i + 1);
        }
    }

    /* free used. */
    return NULL;
}

/********************************************************* MACRO CONFIG FILE */

/* the expanded content of the macro is to be parsed as a configfile_t.
   the following struct stores the content.


   IMPORTANT NOTE:
   ---------------

   in http_config.c there is such a stuff made static,
   which does not implement getch().
   maybe this should be moved to util.c ???
 */
typedef struct {
    int index;               /* current element. */
    int char_index;          /* current char in element. */
    int length;              /* cached length of the current line. */
    array_header * contents; /* array of char * */
} array_lines_t;

/* returns next char or -1.
 */
static int array_getch(void * param)
{
    array_lines_t * ml = (array_lines_t *)param;

    while (ml->char_index >= ml->length) { /* next element */
        if (ml->index >= ml->contents->nelts) {
            return -1;
        }
        ml->index++;
        ml->char_index = 0;
        ml->length = ml->index >= ml->contents->nelts ? 0 :
            strlen(((char **)ml->contents->elts)[ml->index]);
    }

    return (((char **)ml->contents->elts)[ml->index])[ml->char_index++];
}

/* returns a buf a la fgets.
   no more than a line at a time, otherwise the parsing is too much ahead...
   NULL at EOF.
 */
static void * array_getstr(void * buf, size_t bufsize, void * param)
{
    char * buffer = (char *) buf;
    size_t i = 0;
    int next = 0;

    while (i < bufsize - 1 && next != '\n'
           && ((next = array_getch(param)) != -1)) {
        buffer[i++] = (char)next;
    }

    if (next == -1 && i == 0) { /* EOF */
        return NULL;
    }

    buffer[i] = '\0';

    debug(fprintf(stderr, "array_getstr: returning: ---%s---\n", buf));
    return buf;
}

/* close the array stream.
 */
static int array_close(void * param)
{
    array_lines_t * ml = (array_lines_t *)param;
    ml->index = ml->contents->nelts;
    ml->char_index = ml->length;
    return 0;
}

/* this one could be exported.
 */
static configfile_t * make_array_config(ap_pool * p,
                                        array_header * contents,
                                        const char * where)
{
    array_lines_t * ls = (array_lines_t *)ap_palloc(p, sizeof(array_lines_t));

    ls->index      = 0;
    ls->char_index = 0;
    ls->contents   = contents;
    ls->length     = ls->contents->nelts < 1 ? 0 :
        strlen(((char **)ls->contents->elts)[0]);

    return ap_pcfg_open_custom(p, where, (void *)ls,
                               array_getch, array_getstr, array_close);
}


/********************************************************** KEYWORD HANDLING */

/* handles: <Macro macroname macroarg1 macroarg2 ...>
 */
static const char * macro_section(cmd_parms *cmd, void *dummy, const char *arg)
{
    const char * errmsg;
    char ** new, *name, *endp = strrchr(arg, '>');
    macro_t * macro, *old;

    debug(fprintf(stderr, "macro_section -%s-\n", arg));

    macro_init(cmd->temp_pool); /* lazy... */

    if (endp) {
        *endp = '\0';
    }

    /* get name. */
    name = ap_getword_conf(cmd->temp_pool, &arg);

    if (empty_string_p(name)) {
        return "macro definition: name not specified";
    }

    old = get_macro_by_name(all_macros, name);
    if (old) {
        /* already define: warn about the redefinition. */
        ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_WARNING, NULL,
                     "macro '%s' multiply defined.\n"
                     "\t%s, redefined on line %d of %s",
                     old->name, old->location,
                     cmd->config_file->line_number, cmd->config_file->name);
        macro = old;
    }
    else {
        macro = (macro_t *)ap_palloc(cmd->temp_pool, sizeof(macro_t));
    }

    macro->name = name;

    /* get arguments. */
    macro->location = ap_psprintf(cmd->temp_pool,
                                  "defined on line %d of %s",
                                  cmd->config_file->line_number,
                                  cmd->config_file->name);

    macro->arguments = get_arguments(cmd->temp_pool, arg);

    errmsg = check_macro_arguments(cmd->temp_pool, macro);

    if (errmsg) {
        return errmsg;
    }

    /* get contents */
    errmsg = get_lines_till_end_token(cmd->temp_pool, cmd->config_file,
                                      END_MACRO, BEGIN_MACRO, &macro->lines);

    if (errmsg) {
        return ap_psprintf(cmd->temp_pool,
                           "macro '%s' %s\n\tcontents error: %s",
                           macro->name, macro->location, errmsg);
    }

    errmsg = check_macro_contents(cmd->temp_pool, macro);

    if (errmsg) {
        return ap_psprintf(cmd->temp_pool,
                           "macro '%s' %s\n\tcontents checking error: %s",
                           macro->name, macro->location, errmsg);
    }

    macro->processing = 0;

    /* add the new macro. */
    new  = ap_push_array(all_macros);
    *new = (char *)macro;

    return NULL;
}

/* handles: Use name arg1 ...
 */
static const char * use_macro(cmd_parms * cmd, void * dummy, const char * arg)
{
    cmd_parms params;
    char * name, * where;
    const char * errmsg;
    array_header * contents, * replacements;
    macro_t * macro;

    debug(fprintf(stderr, "use_macro -%s-\n", arg));

    macro_init(cmd->temp_pool); /* lazy... */

    name = ap_getword_conf(cmd->temp_pool, &arg);

    if (empty_string_p(name)) {
        return "no macro name specified in " USE_MACRO;
    }

    macro = get_macro_by_name(all_macros, name);

    if (!macro) {
        return ap_psprintf(cmd->temp_pool, "macro '%s' is not defined", name);
    }

    replacements = get_arguments(cmd->temp_pool, arg);

    if (macro->arguments->nelts != replacements->nelts) {
        return ap_psprintf(cmd->temp_pool,
                          "use of macro '%s' %s\n"
                          "\twith %d argument%s instead of %d",
                           macro->name, macro->location, replacements->nelts,
                           replacements->nelts > 1 ? "s" : "", /* grammar;-) */
                           macro->arguments->nelts);
    }

    /* the following is similar to the static process_command_config
       in http_config.h. maybe some common parts should be extracted.
     */
    where = ap_psprintf(cmd->temp_pool,
                        "macro '%s' (%s) used on line %d of %s",
                        macro->name, macro->location,
                        cmd->config_file->line_number, cmd->config_file->name);

    check_macro_use_arguments(where, replacements);

    if (macro->processing) {
        /* tell upwards it is a recursion error. */
        clean_array_of_macros_recursion_flag(all_macros);
        return ap_psprintf(cmd->temp_pool,
                           "%s\n\trecursive use of macro '%s' is invalid.",
                           where, macro->name);
    }

    macro->processing = 1; /* detect recursion. */

    errmsg = process_content(cmd->temp_pool, macro, replacements,
                             NULL, &contents);

    if (errmsg) {
        return ap_psprintf(cmd->temp_pool,
                           "%s\n\terror while substituting:\n%s",
                           where, errmsg);
    }

    params = *cmd; /* copy the full structure. */
    params.end_token = NULL;
    params.config_file = make_array_config(cmd->temp_pool, contents, where);

    errmsg = ap_srm_command_loop(&params, params.server->lookup_defaults);

    if (errmsg) {
        if (!macro->processing) {
            return errmsg; /* not too verbose on macro recursion errors. */
        }
        else {
            return ap_psprintf(cmd->temp_pool,
                               "on line %d of %s\n\texecution error: %s",
                               params.config_file->line_number, where, errmsg);
        }
    }

    if (params.limited != cmd->limited) {
        return ap_psprintf(cmd->temp_pool,
                           "badly nested limits in %s\n", where);
    }

    macro->processing = 0;
    return NULL;
}

/* handles: a lonely </Macro>
 */
static const char * unexpected_end_macro(cmd_parms * dummy1, void * dummy2,
                                         const char * dummy3)
{
    return "unexpected " END_MACRO " encountered";
}

/************************************ ERROR AND WARNING DURING CONFIGURATION */

/* maybe ConfigurationError and ConfigurationWarning could be used?
 */
#define ERROR_KEYWORD "Error"
#define WARNING_KEYWORD "Warning"

/* configuration generated erros or warnings.
 */
static const char * say_it(cmd_parms * parms, void * dummy, const char * arg)
{
    int level = (int) parms->info;

    debug(fprintf(stderr, "level = %d\n", level));

    trim(arg);

    ap_log_error(APLOG_MARK, APLOG_NOERRNO|level, NULL,
                 "on line %d of %s:\n\t%s\n",
                 parms->config_file->line_number,
                 parms->config_file->name,
                 arg);

    if (level & APLOG_ERR) {
        return "configuration file processing aborted by Error.";
    }
    else {
        return NULL;
    }
}

/************************************************************* EXPORT MODULE */

/* macro module commands.
 */
static const command_rec macro_cmds[] =
{
    /* configuration file macro stuff
     */
    { BEGIN_MACRO, macro_section, NULL, OR_ALL, RAW_ARGS,
      "Beginning of a macro definition section." },
    { END_MACRO, unexpected_end_macro, NULL, OR_ALL, NO_ARGS,
      "End of a macro definition section." },
    { USE_MACRO, use_macro, NULL, OR_ALL, RAW_ARGS,
      "Use of a macro." },

    /* configuration errors and warnings.
     */
    { ERROR_KEYWORD, say_it, (void *)APLOG_ERR, OR_ALL, RAW_ARGS,
      "Error in a configuration file." },
    { WARNING_KEYWORD, say_it, (void *)APLOG_WARNING, OR_ALL, RAW_ARGS,
      "Warning in a configuration file." },

    { NULL }
};

/* Module hooks are request-oriented thus it does not suit configuration
   file utils a lot. I haven't found any clean hook to apply something
   before then after configuration file processing. Also what about
   .htaccess files?

   Thus I think that main/http_co*.c would be a better place for this stuff.

   Fabien Coelho (coelho@cri.ensmp.fr).
*/
module MODULE_VAR_EXPORT macro_module =
{
    STANDARD_MODULE_STUFF,
    NULL,                       /* initializer */
    NULL,                       /* create per-dir config */
    NULL,                       /* merge per-dir config */
    NULL,                       /* server config */
    NULL,                       /* merge server config */
    macro_cmds,                 /* command table */
    NULL,                       /* handlers */
    NULL,                       /* filename translation */
    NULL,                       /* check_user_id */
    NULL,                       /* check auth */
    NULL,                       /* check access */
    NULL,                       /* type_checker */
    NULL,                       /* fixups */
    NULL,                       /* logger */
    NULL,                       /* header parser */
    NULL,                       /* child_init */
    NULL,                       /* child_exit */
    NULL                        /* post read-request */
};
